(* This is the target calculus of the sample client. It is a core System F. *)

(* Nominal representation of term variables and binders. *)

(* Nominal or de Bruijn representation of type variables and binders. *)

type ('a, 'b) program = ('a, 'b) definition list

and ('a, 'b) definition =
  | TyDef of tycon * 'a (*'b ?*) list * ('a, 'b) tydefinition
  | ValDef of tevar * ('a, 'b) term

and tevar = string

and tycon = TyConstructor.id

and ('a, 'b) tydefinition =
  | Abbrev of ('a, 'b) typ

and ('a, 'b) term =
  | Var of tevar
  | Abs of tevar * ('a, 'b) typ * ('a, 'b) term
  | App of ('a, 'b) term * ('a, 'b) term
  | Let of tevar * ('a, 'b) term * ('a, 'b) term
  | TyAbs of 'b * ('a, 'b) term
  | TyApp of ('a, 'b) term * ('a, 'b) typ
  | Pair of ('a, 'b) term * ('a, 'b) term
  | Proj of int * ('a, 'b) term
  | Lit of literal

and literal =
  | LInt of int

and ('a, 'b) typ =
  | TyVar of 'a
  | TyForall of 'b * ('a, 'b) typ
  | TyMu of 'b * ('a, 'b) typ
  | TyCon of tycon * ('a, 'b) typ list

type tyvar = int

type nominal_type = (tyvar, tyvar) typ

type debruijn_type =
    (DeBruijn.index, unit) typ

type nominal_term =
  (tyvar, tyvar) term

type nominal_program =
  (tyvar, tyvar) program

type debruijn_term =
    (DeBruijn.index, unit) term

type debruijn_program =
    (DeBruijn.index, unit) program

(* -------------------------------------------------------------------------- *)

(* Constructors. *)

val ftyabs: 'b list -> ('a, 'b) term -> ('a, 'b) term
val ftyapp: ('a, 'b) term -> ('a, 'b) typ list -> ('a, 'b) term

(* -------------------------------------------------------------------------- *)

(* Type-in-type weakening and substitution. *)

(* [lift w k ty] is the type [ty] where the variables at or above index [k]
   are lifted up by [w]. *)

val lift: int -> DeBruijn.index -> debruijn_type -> debruijn_type

(* [subst xty x ty] is the type [ty] where the type variable [x] has been
   replaced with the type [xty]. *)

val subst: debruijn_type -> DeBruijn.index -> debruijn_type -> debruijn_type

(* -------------------------------------------------------------------------- *)

(* Translation of nominal terms to de Bruijn terms. *)

val translate: nominal_term -> debruijn_term

val translate_prog: nominal_program -> debruijn_program
