open Env

(* The unifier will use the following type structure. *)
module S = struct
  open TyConstructor

  type 'a structure =
    | TyCon of TyConstructor.t * 'a list

  let map f t =
    match t with
    | TyCon (t, ts) ->
      TyCon (t, List.map f ts)

  let fold f t accu =
    match t with
    | TyCon (_, ts) ->
      List.fold_right f ts accu

  let iter f t =
    match t with
    | TyCon (_, ts) ->
      List.iter f ts

  exception Iter2

  let convert_abbrev cons vars unify fresh register =

    let tau =
      match TyConstructor.expansion cons with
      | Some t -> t
      | None -> failwith "Not an abbreviation."
    in

    let varenv =
      let c = ref (-1) in
      List.fold_left
        (fun env v ->
           c:= !c + 1; Env.bind !c v env)
        Env.empty  vars in

    let eq = unify in
    let eq' = fun a b ->
      let b' = fresh (Some b) in
      register b';
      eq a b' in
    let ex f =
      let a = fresh None in
      register a; f a
    in

    ex (fun w ->
        convert varenv
        (fun c tys -> TyCon (c, tys))
        tau w eq eq'
        (fun a b -> a ; b ; ())
        ex)


  let iter2 fresh register =

    fun unify t u-> match t, u with
      | TyCon (t, ts), TyCon (u, us) ->
        (* Cas 1 : Two identical constructors -> decomposition *)
        if equal t u
        then decompose t unify ts us
        (* Cas 2 : Two incompatible constructors -> clash *)
        else if incompatible t u
        then raise Iter2
        (* Cas 3 : Expansion*)
        else if rang t >= rang u
        then
          begin
            convert_abbrev t ts unify fresh register
          end
        else raise Iter2



end
