open Env

type id = Type of string
type 't t

val arity : 't t -> int

val name : 't t -> id

val decomposable : 't t -> bool list

val rank : 't t -> int

val expansion : 't t -> 't option

val incompatible : 't t -> 't t -> bool

val equal : 't t -> 't t -> bool

val compare : 't t -> 't t -> int

(* -------------------------------------------------------------------------- *)
(* Functions to compare rank *)

val (>.) : 't t -> 't t -> bool

val (<.) : 't t -> 't t -> bool

val (>=.) : 't t -> 't t -> bool

val (<=.) : 't t -> 't t -> bool

val (=.) : 't t -> 't t -> bool


(* -------------------------------------------------------------------------- *)

type 't env = (id, 't t) Env.t

val empty :
  't env

val bind :
  id -> 't t -> 't env -> 't env

val lookup :
  id -> 't env -> 't t

(* Debug function *)
val cardinal :
  't env -> int

(* -------------------------------------------------------------------------- *)

val decompose : 't t ->
  ('a -> 'b -> unit) ->
  'a list ->
  'b list -> unit

val build_type_cons :
  id -> int -> 't t

val arrow : 't t

val product : 't t

module type LANG = sig
  type tyvar
  type typ

  module TyvarSet : Set.S with type elt = tyvar
  val free_vars : typ env -> typ -> TyvarSet.t

  val is_variable : typ -> tyvar option
  val is_abbrev : typ env -> typ -> (typ t * typ list) option
end

module Make (L : LANG) : sig

  val build_abbrev_cons :
    L.typ env -> id -> L.tyvar list -> L.typ -> L.typ t

    (*
val convert :
    typedecl_env ->
    (ML.tyvar, 'var) Env.t ->
    (t -> 'var list -> 'str) ->
    ML.typ ->
    'var ->
    ('var -> 'var -> 'co) ->
    ('var -> 'str -> 'co) ->
    ('co -> 'co -> 'co) ->
    (('var -> 'co) -> 'co) -> ( 'str option * 'co)
                              *)
end
